@@ -131,7 +131,7 @@ static int check_sysctl (const char *name) {
 
 
 static int tcp_init (const sockaddr_any *dest,
-				unsigned int port_seq, size_t packet_len) {
+			    unsigned int port_seq, size_t *packet_len_p) {
 	int af = dest->sa.sa_family;
 	sockaddr_any src;
 	int mtu;
@@ -305,6 +305,8 @@ static int tcp_init (const sockaddr_any *dest,
 	th->doff = len >> 2;
 
 
+	*packet_len_p = len;
+
 	return 0;
 }
 
@@ -373,52 +375,15 @@ static void tcp_send_probe (probe *pb, int ttl) {
 }
 
 
-static void tcp_recv_probe (int sk, int revents) {
-	int af = dest_addr.sa.sa_family;
-	struct msghdr msg;
-	sockaddr_any from;
-	struct iovec iov;
-	int n, err;
+static probe *tcp_check_reply (int sk, int err, sockaddr_any *from,
+						    char *buf, size_t len) {
 	probe *pb;
-	char buf[1024];
-	char control[1024];
-	struct tcphdr *tcp;
+	struct tcphdr *tcp = (struct tcphdr *) buf;
 	u_int16_t sport, dport;
 
 
-	if (!(revents & (POLLIN | POLLERR)))
-		return;
+	if (len < 8)  return NULL;	    /*  too short   */
 
-	err = !!(revents & POLLERR);
- 
-
-	memset (&msg, 0, sizeof (msg));
-	msg.msg_name = &from;
-	msg.msg_namelen = sizeof (from);
-	msg.msg_control = control;
-	msg.msg_controllen = sizeof (control);
-	iov.iov_base = buf;
-	iov.iov_len = sizeof (buf);
-	msg.msg_iov = &iov;
-	msg.msg_iovlen = 1;
-
-
-	n = recvmsg (raw_sk, &msg, err ? MSG_ERRQUEUE : 0);
-	if (n < 8)	/*  error or too short   */
-		return;
-
-
-	if (af == AF_INET && !err) {	/*  IP header included...  */
-	    struct iphdr *ip = (struct iphdr *) buf;
-	    int hlen = ip->ihl << 2;
-
-	    if (n < hlen + sizeof (struct tcphdr))
-		    return;
-
-	    tcp = (struct tcphdr *) (buf + hlen);
-
-	} else
-	    tcp = (struct tcphdr *) buf;
 
 	if (err) {
 	    sport = tcp->source;
@@ -430,41 +395,33 @@ static void tcp_recv_probe (int sk, int revents) {
 
 
 	if (dport != dest_port)
-		return;
+		return NULL;
 
-	if (!equal_addr (&dest_addr, &from))
-		return;
+	if (!equal_addr (&dest_addr, from))
+		return NULL;
 
 	pb = probe_by_seq (sport);
-	if (!pb)  return;
+	if (!pb)  return NULL;
 
 
-	parse_cmsg (pb, &msg);	    /*   err (if any), tstamp, ttl   */
+	if (!err)  pb->final = 1;
 
-	if (!err) {
-
-	    memcpy (&pb->res, &from, sizeof (pb->res));
-
-	    pb->final = 1;
-	}
+	return pb;
+}
 
 
-	pb->seq = 0;
+static void tcp_recv_probe (int sk, int revents) {
 
-	close (pb->sk);
-	pb->sk = -1;
+	if (!(revents & (POLLIN | POLLERR)))
+		return;
 
-	pb->done = 1;
+	recv_reply (sk, !!(revents & POLLERR), tcp_check_reply);
 }
 
 
 static void tcp_expire_probe (probe *pb) {
 
-	close (pb->sk);
-	pb->sk = -1;
-	pb->seq = 0;
-
-	pb->done = 1;
+	probe_done (pb);
 }
 
 
