@@ -678,10 +678,7 @@ static void print_addr (sockaddr_any *res) {
 	    buf[0] = '\0';
 	    getnameinfo (&res->sa, sizeof (*res), buf, sizeof (buf),
 							    0, 0, NI_IDN);
-	    /*  to be robast...  */
-	    if (!buf[0])  strncpy (buf, str, sizeof (buf));
-
-	    printf (" %s (%s)", buf, str);
+	    printf (" %s (%s)", buf[0] ? buf : str, str);
 	}
 
 	if (as_lookups)
